from sqlalchemy import create_engine
from sqlalchemy.sql import text
from kck.lib.config import kck_config_lookup

db_connections = {}

class KCKDatabase(object):

    @classmethod
    def get_database(cls, dbname):
        global db_connections
        connection_url = kck_config_lookup("databases", dbname, "connection_url")
        if dbname not in db_connections:
            db_connections[dbname] = create_engine(connection_url).connect()
        return db_connections[dbname]

    @classmethod
    def query(cls, dbname, tmpl, params=None):
        dbconx = cls.get_database(dbname)
        stmt = text(tmpl, params)
        if params is not None:
            return dbconx.execute(stmt, params)
        return dbconx.execute(stmt)



import inspect
import os
import pkgutil

import sys

# from kck.lib.kck_data import KCKData
from .kck_database import KCKDatabase
# from .kck_primer import KCKPrimer
from .config import kck_config_lookup
from .framework_actor import FrameworkActor


class KCKUpdater(FrameworkActor):

    name = None
    database = None
    parameters = None
    query_template = None
    data_type = "string"

    @classmethod
    def register_updaters(cls, data_obj):

        # from kck.lib.kck_data import KCKData
        # data_obj = KCKData.get_instance()

        def proc_updater_class(updater_name, cls):
            data_obj.register_updater(updater_name, cls)

        above_dirpath = cls.make_relpath_absolute(
            os.path.join(
                cls.make_relpath_absolute(kck_config_lookup("kck", "updaters_dirpath")),
                '..'
            )
        )
        if above_dirpath not in sys.path:
            sys.path.append(above_dirpath)
        import kupdaters

        for loader, name, ispkg in pkgutil.iter_modules(kupdaters.__path__):
            m = loader.find_module(name).load_module(name)
            for n, c in inspect.getmembers(m, inspect.isclass):
                if not issubclass(c, KCKUpdater) or c.__name__ == "KCKUpdater":
                    continue
                proc_updater_class(name, c())

    def _typed_data(self, d):
        ret = d
        transform = dict(
            string=lambda x: str(x),
            float=lambda x: float(x),
            int=lambda x: int(x)
        )
        if self.data_type in transform:
            return transform[self.data_type](ret)
        return ret

    def update(self, data, queued=False):
        prim_key_param_name = self.primary_key_parameter_name()
        ret = KCKDatabase.query(
            self.database,
            self.query_template["update" if prim_key_param_name in data else "insert"],
            data
        )
        return ret





import datetime
import dateutil
from cassandra.cluster import Cluster, NoHostAvailable
from cassandra.connection import ConnectionException
from dateutil.parser import parser

from kck.lib.exceptions import KCKKeyNotRegistered, KCKKeyNotSetException
from kck.lib.kck_database import KCKDatabase
from kck.lib.kck_primer import KCKPrimer
from kck.lib.kck_updater import KCKUpdater
from kck.lib.yaml_primer import KCKYamlPrimer
from kck.lib.yaml_updater import KCKYamlUpdater
import json
from .config import kck_config_lookup


class CassandraActor(object):
    prim_cache_name = None
    sec_cache_name = None
    update_queue_name = None
    refresh_queue_name = None
    refresh_counter_name = None
    cluster = None
    session = None
    cache_obj = None
    data_obj = None
    updaters = None
    primers = None
    keysep = "/"

    def init(self, inhibit_framework_registration=False, cache_obj=None, data_obj=None):
        try:
            self.session = self.cluster.connect(kck_config_lookup("cassandra", "keyspace"))
        except (ConnectionException, NoHostAvailable):
            self.cluster.shutdown()
            self._cache_create_keyspace(kck_config_lookup("cassandra", "keyspace"))
            self.cluster = Cluster(kck_config_lookup("cassandra", "hosts"))
            self.session = self.cluster.connect(kck_config_lookup("cassandra", "keyspace"))
        self._cache_create_table(self.prim_cache_name)
        self._cache_create_table(self.sec_cache_name)
        self._queued_updates_create_table(self.update_queue_name)
        self._queued_refreshes_create_tables(self.refresh_queue_name)

        self.cache_obj = cache_obj
        self.data_obj = data_obj

        if not inhibit_framework_registration:
            self._register_updaters()
            self._register_primers()

        self._register_hooks()

    def primitive_init(self):
        self.prim_cache_name = kck_config_lookup("cassandra", "tables", "primary_cache")
        self.sec_cache_name = kck_config_lookup("cassandra", "tables", "secondary_cache")
        self.update_queue_name = kck_config_lookup("cassandra", "tables", "queued_updates")
        self.refresh_queue_name = kck_config_lookup("cassandra", "tables", "queued_refreshes")
        self.refresh_counter_name = kck_config_lookup("cassandra", "tables", "queued_refreshes_counter")
        self.cluster = Cluster(kck_config_lookup("cassandra", "hosts"))

        self.updaters = {}
        self.primers = {}

    def primer_obj(self, key):
        try:
            return self.cache_obj.primers[key.split(self.keysep)[0]]
        except KeyError:
            raise KCKKeyNotRegistered(key)

    def updater_obj(self, key):
        return self.data_obj.updaters[key.split(self.keysep)[0]]

    def do_processes(self):

        type_map = dict(
            sql=self._sql_process,
            key=self._key_process
        )

        for key, actor_obj in self.cache_obj.primers.items():
            try:
                processes = actor_obj.processes
            except AttributeError:
                continue

            for process_name, process_dict in processes.items():
                print("process for key: {}, name: {}".format(key, process_name))
                if process_dict["type"] in type_map:
                    type_map[process_dict["type"]](
                        process_name=process_name,
                        process_dict=process_dict,
                        actor_obj=actor_obj
                    )

    def queued_refresh_data(self, query=None):
        # # get queue names from config
        # refresh_queue_name = kck_config_lookup(
        #     "cassandra", "tables", "queued_refreshes"
        # )
        # refresh_counter_name = kck_config_lookup(
        #     "cassandra", "tables", "queued_refreshes_counter"
        # )

        # async refresh queue query
        q_hints = "select kck_key, kck_hints from {}".format(self.refresh_queue_name)
        pq_hints = self.session.prepare(q_hints)
        fut_hints = self.session.execute_async(pq_hints)

        # async refresh counters query
        q_counts = "select kck_key, refresh_request_counter from {}".format(
            self.refresh_counter_name
        )
        pq_counts = self.session.prepare(q_counts)
        fut_counts = self.session.execute_async(pq_counts)

        queue_dict = {}
        for key_hints_tuple in fut_hints.result():
            key, hints = key_hints_tuple
            queue_dict[key] = {"hints": hints}

        for key_count_tuple in fut_counts.result():
            key, cnt = key_count_tuple
            if key not in queue_dict:
                queue_dict[key] = dict(count=cnt)
            else:
                queue_dict[key]["count"] = cnt

        return queue_dict

    def queue_refresh_count(self, key):
        q = "select refresh_request_counter from {} where kck_key= ?".format(self.refresh_counter_name)
        pq = self.session.prepare(q)
        future = self.session.execute_async(pq)
        ret = future.result()[0]
        return ret

    def perform_queued_refreshes(self, query=None):

        refresh_data = self.queued_refresh_data(query)

        for key, refresh_request_entry in refresh_data.items():
            hints = None
            if "hints" in refresh_request_entry and \
                    refresh_request_entry["hints"] is not None:
                hints = refresh_request_entry["hints"]
            self.cache_obj.refresh(key, hints)

    def raw_queue_update(self, key, data):
        q = "insert into {} (kck_key, kck_data) values (?, ?)".format(self.update_queue_name)
        pq = self.session.prepare(q)
        future = self.session.execute_async(pq, (key, data))
        return future.result()

    def raw_cache_put(self, key, tbl, val, version=None, check_last_version=None, expire=None, modified=None):
        """sets the value associated with <key> in the table <tbl>.
           if callback_func is defined, the query is executed asynchronously
           and the future resulting from the call to execute_async method"""

        new_version = version if version is not None else self.gen_new_version()

        ttl_clause = "" if expire is None else "using ttl {} ".format(expire.seconds)
        current_timestamp = datetime.datetime.utcnow() if modified == None else modified

        # --- if <check_last_version> is None, then don't do the version check
        if check_last_version is None:

            upsert_query = "update {} {}set kck_value = ?, version = ?, modified = ? where kck_key = ?".format(tbl, ttl_clause)
            prepared_query = self.session.prepare(upsert_query)
            future = self.session.execute_async(prepared_query, (val, new_version, current_timestamp, key))
            return future.result()

        # --- update only if version = <check_last_version>
        upsert_query = "update {} {}set kck_value = ?, version = ?, modified = ? where kck_key = ? if version = ?".format(tbl, ttl_clause)
        prepared_query = self.session.prepare(upsert_query)

        future = self.session.execute_async(prepared_query, (val, new_version, current_timestamp, key, check_last_version))

        return future.result()

    def raw_cache_delete(self, key, tbl):
        delete_query_str = "delete from {} where kck_key = ?".format(tbl)

        prepared_query = self.session.prepare(delete_query_str)

        future = self.session.execute_async(prepared_query, (key,))
        return future.result()

    def raw_cache_get(self, key, tbl):
        """looks up the value associated with <key> in the table <tbl>.
           if callback_func is defined, the query is executed asynchronously
           and the future resulting from the call to execute_async method"""
        prepared_query = self.session.prepare(
            "select kck_value, version, modified from {} where kck_key = ?".format(
                tbl
            )
        )
        future = self.session.execute_async(prepared_query, (key,))

        ret =  future.result()
        return ret

    def raw_queue_refresh(self, key, hints=None):

        # increment refresh counter
        q_inc = "update {} set refresh_request_counter = refresh_request_counter + 1 where kck_key = ?".format(self.refresh_counter_name, key)
        pq_inc = self.session.prepare(q_inc)
        future_inc = self.session.execute_async(pq_inc, (key,))

        # if there's no hints, incrementing the refresh counter is good enough
        if hints is None:
            ret =  future_inc.result()
            return ret

        # update the refresh table with hints
        q_hints = "update {} set kck_hints=kck_hints".format(self.refresh_queue_name) + \
                  ' + ' + '{' + "'" + self._to_encoded_json(hints) + "'" + '}' + \
                  " where kck_key = ?"
        pq_hints = self.session.prepare(q_hints)
        if hints is not None:
            if "modified" in hints:
                hints["modified"] = str(hints["modified"])
        future_hints = self.session.execute_async(pq_hints, (key,))

        # return
        ret = future_hints.result()
        return ret

    def _to_encoded_json(self, d):
        def serialize_datetime(obj):
            if isinstance(obj, (datetime.datetime, datetime.date)):
                return obj.isoformat()
            raise TypeError("Type %s not serializable" % type(obj))
        return json.dumps(d, default=serialize_datetime)


    def _key_process(self, process_name, process_dict, actor_obj):

        if "hooks" in process_dict:

            hooks = process_dict["hooks"]

            if "pre" in hooks and hooks["pre"] is not None:
                for pre_hook in hooks["pre"]:
                    pre_hook(self)

            if "value" in hooks and hooks["value"] is not None:
                for value_hook in hooks["value"]:
                    if "key_attribute" in process_dict:
                        key_ptr = getattr(actor_obj, process_dict["key_attribute"])
                        if callable(key_ptr):
                            key = key_ptr()
                        else:
                            key = key_ptr

                    if key is not None:
                        try:
                            value_hook(actor_obj, self.cache_obj.get(key)["value"])
                        except KCKKeyNotSetException:
                            pass

    def _sql_process(self, process_name, process_dict, actor_obj):

        if "hooks" in process_dict:

            hooks = process_dict["hooks"]

            if "pre" in hooks and hooks["pre"] is not None:
                for pre_hook in hooks["pre"]:
                    pre_hook(self)

            if "row" in hooks and hooks["row"] is not None:
                rowset = KCKDatabase.query(
                    process_dict["database"],
                    process_dict["query"]
                )
                for row in rowset:
                    for row_hook in hooks["row"]:
                        row_hook(self, row)

            if "post" in hooks and hooks["post"] is not None:
                for post_hook in hooks["post"]:
                    post_hook(self)

    def _register_primers(self):
        """registers yaml and class-based primers"""
        KCKPrimer.register_primers(self.cache_obj)
        KCKYamlPrimer.register_primers(self.cache_obj)

    def _register_updaters(self):
        """registers yaml and class-based updaters"""
        KCKYamlUpdater.register_updaters(self.data_obj)
        KCKUpdater.register_updaters(self.data_obj)

    def _register_hooks(self):

        # register primer dependencies
        for _, p in self.primers.items():
            p.register_hooks()

        # register updater dependencies
        for _, u in self.updaters.items():
            u.register_hooks()

    def _cache_create_table(self, tbl):
        ks = kck_config_lookup("cassandra", "keyspace")
        self.session.execute("""
            CREATE TABLE IF NOT EXISTS {}.{} (
              kck_key text PRIMARY KEY,
              kck_value blob,
              version int,
              modified timestamp,
            )
        """.format(ks, tbl))

    def _queued_updates_create_table(self, tbl):
        ks = kck_config_lookup("cassandra", "keyspace")
        self.session.execute("""
            CREATE TABLE IF NOT EXISTS {}.{} (
              kck_key text PRIMARY KEY,
              kck_data blob,
              version int
            )
        """.format(ks, tbl))

    # counter explanation: https: // stackoverflow.com / questions / 23145817 / update - a - cassandra - integer - column - using - cql
    def _queued_refreshes_create_tables(self, tbl):
        ks = kck_config_lookup("cassandra", "keyspace")

        self.session.execute("""
            CREATE TABLE IF NOT EXISTS {}.{} (
              kck_key text PRIMARY KEY,
              kck_hints set<text>,
              version int
            )
        """.format(ks, tbl))

        self.session.execute("""
            CREATE TABLE IF NOT EXISTS {}.{} (
              kck_key text PRIMARY KEY,
              refresh_request_counter counter
            )
        """.format(ks, tbl + "_counter"))

    @staticmethod
    def _cache_create_keyspace(keyspace_name):
        cluster = Cluster(kck_config_lookup("cassandra", "hosts"))
        create_keyspace_session = cluster.connect()
        create_keyspace_session.execute("""
            CREATE KEYSPACE IF NOT EXISTS {}
            WITH REPLICATION = {{
               'class' : 'SimpleStrategy',
               'replication_factor' : 1
            }}
        """.format(keyspace_name))
        create_keyspace_session.shutdown()
        cluster.shutdown()




import os
import random
import zlib
import pickle

import datetime

from .config import kck_config_lookup
from .exceptions import PrimerComputerReturnedNoResults, KCKKeyNotSetException, KCKKeyNotRegistered, KCKUnknownKey
from .cassandra_actor import CassandraActor

SCRIPT_DIRPATH = os.path.dirname(os.path.realpath(__file__))
KCK_SOURCES_DIRPATH = os.path.join(SCRIPT_DIRPATH, "..")

process_cache_singleton = None


class KCKCache(CassandraActor):

    prime_on_cache_miss = False
    tpickle = False
    tcompress = False

    def __init__(
        self,
        inhibit_framework_registration=False,
        key_string_sep="/",
        serialize=True,
        compress=True,
        data_obj=None
    ):
        self.primitive_init()
        if data_obj is None:
            from .kck_data import KCKData
            self.init(
                inhibit_framework_registration=inhibit_framework_registration,
                cache_obj=self,
                data_obj=KCKData.get_instance(cache_obj=self)
            )
        else:
            self.init(
                inhibit_framework_registration=inhibit_framework_registration,
                cache_obj=self,
                data_obj=data_obj
            )

        self.prime_on_cache_miss = kck_config_lookup("kck", "prime_on_cache_miss")
        self.keysep = key_string_sep

        self.tpickle = serialize
        self.tcompress = compress

    @classmethod
    def get_instance(cls, **kwargs):
        global process_cache_singleton
        if process_cache_singleton is None:
            process_cache_singleton = cls(**kwargs)
        return process_cache_singleton

    @staticmethod
    def gen_new_version():
        """returns a new random integer to be used for versioning"""
        minver = -2147483648
        maxver = 2147483647
        return random.randint(minver, maxver)

    def get(self, key):
        """given a key, try to find a corresponding value in the primary cache. failing that,
           try to find the corresponding value in the secondary cache. failing that, if
           self.prime_on_cache_miss is True, call the primer"""
        try:
            return self._primcache_get(key)

        except KCKKeyNotSetException:

            # if kck is depending on the val cached in the
            # secondary cache, it's time to refresh
            # self.refresh(key, queued=True)

            if not self.prime_on_cache_miss:

                ret = self._seccache_get(key)
                self.refresh(key, queued=True)
                return ret

            try:
                ret = self._seccache_get(key)
                self.refresh(key, queued=True)
                return ret

            except KCKKeyNotSetException:
                try:
                    return self._prime(key)
                except PrimerComputerReturnedNoResults:
                    raise KCKKeyNotSetException(key, self.prim_cache_name, msg="primer found, but compute() returned no results")


    def refresh(self, key, hints=None, queued=False):

        if bool(queued):

            if hints is not None and type(hints) is dict:
                hints["modified"] = datetime.datetime.utcnow()

            return self.raw_queue_refresh(key=key, hints=hints)

        print("refreshing key: {}".format(key))
        ret =  self._prime(key, hints=hints)
        return ret

    def primer_registered(self, name):
        return bool(name in self.primers)

    def register_primer(self, name, primer_obj):
        print("registering primer: {}".format(name))
        primer_obj.set_cache(self)
        primer_obj.set_data(self.data_obj)
        primer_obj.set_name(name)
        self.primers[name] = primer_obj



    def _prime(self, key, hints=None):

        try:
            p = self.primer_obj(key)

            p.do_hooks("pre_prime", key=key)

            primed_val = p.compute(key)

        except KCKKeyNotRegistered:
            raise KCKUnknownKey(key)

        ret = self._primcache_put(key, primed_val, p.expire)
        self._seccache_put(key, primed_val, p.expire)

        p.do_hooks("post_prime", key=key, hints=ret)

        return ret

    def _primcache_get(self, key):
        return self._cache_get(key, self.prim_cache_name)

    def _seccache_get(self, key):
        return self._cache_get(key, self.sec_cache_name)

    def _primcache_put(self, key, val, expire=None):
        return self._cache_put(key, self.prim_cache_name, val, expire=expire)

    def _seccache_put(self, key, val, expire=None):
        return self._cache_put(key, self.sec_cache_name, val, expire=expire)

    def _cache_get(self, key, tbl):

        def decomprickle(val):
            ret = val
            # print("decom ret[{}]: {}".format(type(ret), ret))
            if self.tcompress:
                ret = zlib.decompress(ret)
            if self.tpickle:
                ret = pickle.loads(ret)["d"]
            return ret

        raw_get_result = self.raw_cache_get(key, tbl)

        try:
            return dict(success=True, version=raw_get_result[0][1], key=key, value=decomprickle(raw_get_result[0][0]), tbl=tbl, modified=raw_get_result[0][2])
        except IndexError:
            raise KCKKeyNotSetException(key, tbl)

    def _cache_put(self, key, tbl, val, version=None, check_last_version=None, expire=None):
        
        def comprickle(v):
            nv = v
            if self.tpickle:
                nv = pickle.dumps({"d": nv})
            if self.tcompress:
                nv = zlib.compress(nv)
            return nv

        new_version = version if version is not None else self.gen_new_version()

        newval = comprickle(val)

        current_timestamp = datetime.datetime.utcnow()

        self.raw_cache_put(key, tbl, newval, version=new_version, check_last_version=check_last_version, expire=expire, modified=current_timestamp)
        ret = dict(success=True, version=new_version, key=key, value=val, tbl=tbl, modified=current_timestamp)
        return ret

    def _cache_delete(self, key, tbl):
        self.raw_cache_delete(key, tbl)
        return dict(success=True, key=key, tbl=tbl)




from .log import log

def log_error(exc, msg=None):
    if msg is None:
        log.error("Operation failed: %s", exc)
    else:
        log.error("Operation failed: %s (%s)", msg, exc)



def handle_error(exc, msg, on_error=None):
    if on_error is not None:
        on_error(exc)
    log_error(exc)


import yaml

from .err import handle_error

class YAMLReader(object):
    @staticmethod
    def read_yaml(filepath, on_error=None):
        """read/parse a yaml file, return resulting object"""
        with open(filepath, 'r') as stream:
            try:
                ret = yaml.load(stream)
                return ret
            except yaml.YAMLError as exc:
                handle_error(exc, "unable to read yaml file: {}".format(filepath), on_error)


import os
import yaml
from .err import handle_error, log_error

SCRIPT_DIRPATH = os.path.dirname(os.path.realpath(__file__))
KCK_SOURCES_DIRPATH = os.path.join(SCRIPT_DIRPATH, "..")

class NotEnoughParameters(Exception):
    pass

class FrameworkActor(object):

    keysep = "/"
    parameters = []
    cache_obj = None
    data_obj = None
    hooks = None
    name = None

    @staticmethod
    def make_relpath_absolute(dirpath):
        """given any dirpath, relative or absolute, return an absolute dirpath"""
        if dirpath[0] == '/':
            return dirpath
        return os.path.abspath(os.path.join(KCK_SOURCES_DIRPATH, dirpath))

    @staticmethod
    def proc_files_in_dirpath(dirpath, filter_func, callback_func):
        """walk all files in dirpath, recursing into subdirs, and, if filter_func
           returns True for a given filepath, call callback_func with the filepath
           as a parameter"""
        for subdir, dirs, files in os.walk(dirpath):
            for file in files:
                fp = os.path.join(subdir, file)
                if filter_func(fp):
                    callback_func(os.path.join(subdir, file))

    @staticmethod
    def read_yaml(filepath, on_error=None):
        """read/parse a yaml file, return resulting object"""
        with open(filepath, 'r') as stream:
            try:
                return yaml.load(stream)
            except yaml.YAMLError as exc:
                handle_error(exc, "unable to init config object", on_error)

    def set_name(self, name_str):
        self.name = name_str

    def get_name(self):
        return self.name

    def set_cache(self, cache_obj):
        self.cache_obj = cache_obj

    def get_cache(self):
        return self.cache_obj

    def set_data(self, data_obj):
        self.data_obj = data_obj

    def get_data(self):
        return self.data_obj

    def primary_key_parameter_name(self):
        for p in self.parameters:
            if "primary_key" in p and bool(p["primary_key"]):
                return p["name"]
        return None

    def dependency_key(self):
        ret = self.name
        for p in self.parameters:
            ret += self.keysep + ":{}".format(p["name"])
        return ret

    def dependency_key_to_param_dict(self, key_tmpl, key):
        print("dependency_key_to_param_dict entered")
        key_tmpl_names_with_colons = \
            key_tmpl.split(self.keysep) if self.keysep in key_tmpl else [key_tmpl]
        print("key_tmpl_names_with_colons: {}".format(key_tmpl_names_with_colons))
        key_tmpl_names = \
            [ x.split(':')[1] if ':' in x else x for x in key_tmpl_names_with_colons][1:]
        print("key_tmpl_names: {}".format(key_tmpl_names))
        key_values = key.split(self.keysep)[1:] if self.keysep in key else [key]
        print("key_values: {}".format(key_values))

        ret = {}
        for ndx, name in enumerate(key_tmpl_names):
            ret[name] = key_values[ndx]

        return ret

    def key_to_param_dict(self, key_str):

        transform = dict(
            string=lambda x: str(x),
            float=lambda x: float(x),
            int=lambda x: int(x)
        )

        if key_str is None or key_str == "":
            return {}

        param_list = key_str.split(self.keysep)[1:] if self.keysep in key_str else []
        # print("key_str: {}, param_list: {}".format(key_str, param_list))
        try:
            if len(param_list) < len(self.parameters):
                raise NotEnoughParameters
            ret = {}
            for ndx, param_dict in enumerate(self.parameters):
                ret[param_dict["name"]] = param_list[ndx]
                if "type" in param_dict:

                    try:
                        ret[param_dict["name"]] = transform[param_dict["type"]](ret[param_dict["name"]])
                    except ValueError:
                        ret[param_dict["name"]] = None

            return ret
        except (IndexError, NotEnoughParameters):

            # this builds the dict when there's no primary key (as when records
            #  are being created)

            if len(param_list) == 0:
                return {}

            # if there's just one arg, it's the primary key and we're updating
            if len(param_list) == 1:
                primary_key_name = None
                for ndx, param_dict in enumerate(self.parameters):
                    if "primary_key" in param_dict and bool(param_dict["primary_key"]):
                        primary_key_name = param_dict["name"]
                        return { primary_key_name: param_list[0] }

            ret = {}
            offset = 0
            for ndx, param_dict in enumerate(self.parameters):
                if "primary_key" in param_dict and bool(param_dict["primary_key"]):
                    offset = -1
                    continue
                print("ndx: {}".format(ndx))
                ret[param_dict["name"]] = param_list[ndx+offset]
                if "type" in param_dict:
                    try:
                        ret[param_dict["name"]] = transform[param_dict["type"]](ret[param_dict["name"]])
                    except ValueError:
                        ret[param_dict["name"]] = None
            return ret

    def param_dict_to_key(self, param_dict):
        key_components = []
        for p in self.parameters:
            try:
                key_components.append(str(param_dict[p["name"]]))
            except KeyError:
                raise NotEnoughParameters
        return self.keysep.join(key_components)

    def register_hooks(self):
        if self.hooks is None:
            self.hooks = {}

    def add_hook(self, event, keystr, func):
        if self.hooks is None:
            self.hooks = {}
        if event not in self.hooks:
            self.hooks[event] = []
        self.hooks[event].append((keystr, func))

    def do_hooks(self, event, *args, **kwargs):

        if self.hooks is None:
            self.hooks = {}

        print("do_hooks({}) for {}".format(event, self.name))
        try:
            for hook_tuple in self.hooks[event]:
                keystr, hfunc = hook_tuple
                print("  found {} hook named {} with key: {}...".format(event, hfunc.__name__, keystr))
                kwargs["param_key"] = keystr
                try:
                    hfunc(*args, **kwargs)
                except Exception as e:
                    print("  err: {}".format(e))
                    log_error(e)
        except KeyError:
            pass
